# SPDX-License-Identifier: BSD-2-Clause
""" Compare two specification sets. """

# Copyright (C) 2019, 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import dataclasses
import os
import re
import tempfile
from typing import Any

from specitems import (Item, ItemCache, ItemCacheConfig, ItemGetValueContext,
                       ItemValueProvider, ItemSelection, TextContent,
                       TextMapper)
from specware import run_command, validate

from .directorystate import DirectoryState
from .sphinxbuilder import SphinxBuilder

_COMMIT = re.compile(r"commit ([0-9a-f]+)$")


@dataclasses.dataclass
class CompareSpecsConfig:
    # pylint: disable=too-many-instance-attributes
    """ Represents a specification comparison configuration. """
    repository: DirectoryState
    root_uid: str
    current_revision: str
    previous_revision: str
    ignored_commits: list[str]
    spec_paths: list[str]
    selection: ItemSelection
    enabled_set_actions: list[dict[str, Any]]
    label: str


class _LogParser:
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes

    def __init__(self, uids: set[str], ignored_commits: list[str],
                 spec_paths: list[str]):
        self._uids = uids
        self._ignored_commits = ignored_commits
        self._spec_paths = sorted(spec_paths, reverse=True)
        self._commits: list[list[str]] = []
        self._current_commit: dict[str, Any] = {}
        self._current_chunk: list[str] = []
        self._current_diff: dict[str, Any] = {}
        self.consume = self._expect_commit

    def _get_diff_uid(self, diff: dict, which: str) -> str | None:
        if diff[which] == "/dev/null":
            return None
        path = diff[which][2:]
        for prefix in self._spec_paths:
            uid = path.removeprefix(prefix)
            if path != uid and path.endswith(".yml"):
                return uid[:-4]
        raise ValueError

    def _finalize_diff(self):
        diff = self._current_diff
        self._current_diff = {}
        if diff:
            try:
                a_uid = self._get_diff_uid(diff, "a")
            except ValueError:
                return
            try:
                b_uid = self._get_diff_uid(diff, "b")
            except ValueError:
                return
            if b_uid is not None and b_uid not in self._uids:
                return
            diff["a-uid"] = a_uid
            diff["b-uid"] = b_uid
            self._current_commit["diffs"].append(diff)

    def _finalize(self):
        if self._current_chunk:
            self._current_diff["chunks"].append(self._current_chunk)
            self._current_chunk = []
        self._finalize_diff()
        commit = self._current_commit
        self._current_commit = {}
        if commit and commit["diffs"] and commit[
                "commit"] not in self._ignored_commits:
            self._commits.append(commit)

    def finalize(self):
        """ Finalizes the parsing and returns the commits. """
        self._finalize()
        return self._commits

    def _expect_commit(self, line: str) -> None:
        mobj = _COMMIT.match(line)
        assert mobj
        self._finalize()
        self._current_commit = {
            "diffs": [],
            "commit": mobj.group(1),
            "message": []
        }
        self.consume = self._expect_author

    def _expect_author(self, line: str) -> None:
        key = "Author: "
        assert line.startswith(key)
        self._current_commit["author"] = line[len(key):]
        self.consume = self._expect_date

    def _expect_date(self, line: str) -> None:
        key = "Date: "
        assert line.startswith(key)
        self._current_commit["date"] = line[len(key):]
        self.consume = self._expect_empty_line

    def _expect_empty_line(self, line: str) -> None:
        assert not line
        self.consume = self._message

    def _message(self, line: str) -> None:
        space = "    "
        if line.startswith(space):
            self._current_commit["message"].append(line[len(space):])
            return
        self.consume = self._diff_start
        self._diff_start(line)

    def _diff_start(self, line: str) -> None:
        self._finalize_diff()
        if line.startswith("---"):
            self._current_diff = {"a": line[4:], "chunks": []}
            self.consume = self._expect_diff_plus

    def _expect_diff_plus(self, line: str) -> None:
        assert line.startswith("+++")
        self._current_diff["b"] = line[4:]
        self.consume = self._expect_chunk_start

    def _expect_chunk_start(self, line: str) -> None:
        assert line.startswith("@@")
        self._current_chunk = [line]
        self.consume = self._chunk

    def _chunk(self, line: str) -> None:
        if line.startswith((" ", "-", "+")):
            self._current_chunk.append(line)
            return
        self._current_diff["chunks"].append(self._current_chunk)
        self._current_chunk = []
        if line.startswith("@@"):
            self.consume = self._expect_chunk_start
            self._expect_chunk_start(line)
            return
        if line.startswith("commit"):
            self.consume = self._expect_commit
            self._expect_commit(line)
            return
        if line.startswith("diff"):
            self.consume = self._diff_start
            self._diff_start(line)
            return


def _validate(_item: Item, validated: bool) -> bool:
    return validated


def _get_uids_current(config: CompareSpecsConfig) -> set[str]:
    return set(item.uid for item in validate(
        config.repository.item.cache[config.root_uid], _validate))


def _get_uids_previous(config: CompareSpecsConfig) -> set[str]:
    with tempfile.TemporaryDirectory() as tmp_dir:
        cache_config = ItemCacheConfig(paths=[
            os.path.join(config.repository.directory, path)
            for path in config.spec_paths
        ],
                                       cache_directory=tmp_dir)
        item_cache = ItemCache(cache_config)
        selection = config.selection.clone(item_cache)
        for action in config.enabled_set_actions:
            selection.apply_action(action)
        item_cache.set_selection(selection)
        uids: set[str] = set()
        for item in item_cache.values():
            if not item.enabled or item.type == "spec":
                continue
            uids.add(item.uid)
        return uids


def _get_commits(config: CompareSpecsConfig) -> list[str]:
    stdout: list[str] = []
    status = run_command([
        "git", "log", "--no-renames", "-p",
        f"{config.previous_revision}..{config.current_revision}"
    ] + config.spec_paths, config.repository.directory, stdout)
    assert status == 0
    return stdout


def compare_specs(content: TextContent, config: CompareSpecsConfig) -> None:
    """ Compare two specifications according to the configuration. """
    uids_current = _get_uids_current(config)
    uids_previous = _get_uids_previous(config)
    all_uids = uids_current.union(uids_previous)
    parser = _LogParser(all_uids, config.ignored_commits, config.spec_paths)
    for line in _get_commits(config):
        parser.consume(line.rstrip("\n"))
    with content.label_scope(config.label):
        for commit in reversed(parser.finalize()):
            with content.section(commit["message"][0]):
                content.add(commit["message"][2:])
                for diff in commit["diffs"]:
                    a_uid = diff["a-uid"]
                    b_uid = diff["b-uid"]
                    if a_uid is None:
                        with content.section(f"spec:{b_uid}"):
                            content.add(
                                "This change added the specification item.")
                        continue
                    if b_uid is None:
                        with content.section(f"spec:{a_uid}"):
                            content.add(
                                "This change removed the specification item.")
                        continue
                    with content.section(f"spec:{b_uid}"):
                        for chunk in diff["chunks"]:
                            for begin in range(0, len(chunk), 100):
                                with content.directive("code-block", "diff"):
                                    content.add(chunk[begin:begin + 100])


class CompareSpecsProvider(ItemValueProvider):
    """ Provides a specification comparison. """

    # pylint: disable=too-few-public-methods

    def __init__(self, builder: SphinxBuilder) -> None:
        super().__init__(builder.mapper)
        self._builder = builder
        self.mapper.add_get_value(f"{builder.item.type}:/compare-specs",
                                  self._get_compare_specs)

    def _get_compare_specs(self, ctx: ItemGetValueContext) -> str:
        builder = self._builder
        with builder.section_level_scope(ctx) as key:
            assert key
            link, repo = builder.input_link_by_key("spec-compare",
                                                   "spec-compare-key", key)
            assert isinstance(repo, DirectoryState)
            data = builder.substitute(link.data)
            config = CompareSpecsConfig(
                repository=repo,
                root_uid=data["root-uid"],
                current_revision=data["current-revision"],
                previous_revision=data["previous-revision"],
                ignored_commits=data["ignored-commits"],
                spec_paths=data["spec-paths"],
                selection=builder.director.item_cache.active_selection,
                enabled_set_actions=data["enabled-set-actions"],
                label=data["label"])
            assert isinstance(self.mapper, TextMapper)
            content = self.mapper.create_content(
                section_level=builder.section_level)
            compare_specs(content, config)
            return "\n".join(content)



# SPDX-License-Identifier: BSD-2-Clause
""" Provides a command line interface to update a performance limits item. """

# Copyright (C) 2023 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import contextlib
import json
import sys
from typing import Any, Dict, List

from specitems import (Item, ItemCache, ItemCacheConfig,
                       create_argument_parser, create_config, init_logging,
                       item_is_enabled)
from specware import load_specware_config, SpecWareTypeProvider

from .testoutputparser import augment_report

_LimitsByUID = Dict[str, Dict[str, Dict[str, float]]]


def _calculate_limits(runtime_measurement: Dict[str, Any], uid: str,
                      variant: str, limits_by_uid: _LimitsByUID,
                      args: Any) -> Dict[str, float]:
    minimum = runtime_measurement["min"]
    median = runtime_measurement["q2"]
    maximum = runtime_measurement["max"]
    min_lower_bound = args.lower_bound_scaler * minimum
    max_upper_bound = args.upper_bound_scaler * maximum
    median_lower_bound = max(args.lower_bound_scaler * median, min_lower_bound)
    median_upper_bound = min(args.upper_bound_scaler * median, max_upper_bound)
    if args.lazy:
        try:
            current_limits = limits_by_uid[uid][variant]
        except KeyError:
            pass
        else:
            limit = current_limits["min-lower-bound"]
            if limit <= minimum:
                min_lower_bound = limit
            limit = current_limits["median-lower-bound"]
            if limit <= median:
                median_lower_bound = limit
            limit = current_limits["median-upper-bound"]
            if limit >= median:
                median_upper_bound = limit
            limit = current_limits["max-upper-bound"]
            if limit >= maximum:
                max_upper_bound = limit
    return {
        "max-upper-bound": max_upper_bound,
        "median-lower-bound": median_lower_bound,
        "median-upper-bound": median_upper_bound,
        "min-lower-bound": min_lower_bound
    }


def _update_perf_limits(reports: List[Any], name_to_uid: Dict[str, str],
                        limits_by_uid: _LimitsByUID, args: Any) -> None:
    for report in reports:
        if "performance" in report["executable"]:
            if "info" not in report:
                augment_report(report, report["output"])
            try:
                test_cases = report["test-suite"]["test-cases"]
            except KeyError:
                print("no performance data available for:",
                      report["executable"])
                continue
            for test_case in test_cases:
                for runtime_measurement in test_case["runtime-measurements"]:
                    uid = name_to_uid[runtime_measurement["name"]]
                    variant = runtime_measurement["variant"]
                    limits_by_uid.setdefault(uid,
                                             {})[variant] = _calculate_limits(
                                                 runtime_measurement, uid,
                                                 variant, limits_by_uid, args)


def _write_perf_limits(perf_limits: Item, limits_by_uid: _LimitsByUID) -> None:
    links = [
        link for link in perf_limits["links"]
        if link["role"] != "performance-runtime-limits"
    ]
    for uid, limits in sorted(limits_by_uid.items()):
        links.append({
            "limits": limits,
            "role": "performance-runtime-limits",
            "uid": uid
        })
    perf_limits.data["links"] = links
    perf_limits.save()


def cliupdateperf(argv: list[str] = sys.argv) -> None:
    """ Updates a performance limits item. """
    parser = create_argument_parser()
    parser.add_argument("--config-file",
                        type=str,
                        default=None,
                        help="use this configuration file")
    parser.add_argument('--lazy',
                        action="store_true",
                        help="only update necessary limits")
    parser.add_argument('--lower-bound-scaler',
                        type=float,
                        default=0.9,
                        help="scaler to define lower bounds (default: 0.9)")
    parser.add_argument('--upper-bound-scaler',
                        type=float,
                        default=1.1,
                        help="scaler to define upper bounds (default: 1.1)")
    parser.add_argument(
        "perf_limits_uid",
        metavar="PERF_LIMITS_UID",
        nargs=1,
        help="the UID of the performance limits item to update")
    parser.add_argument("reports",
                        metavar="REPORTS",
                        nargs="+",
                        help="the test report files")
    args = parser.parse_args(argv[1:])
    init_logging(args)
    config, working_directory = load_specware_config(args.config_file)
    with contextlib.chdir(working_directory):
        item_cache = ItemCache(create_config(config["spec"], ItemCacheConfig),
                               type_provider=SpecWareTypeProvider({}),
                               is_item_enabled=item_is_enabled)
        name_to_uid: Dict[str, str] = {}
        for item in item_cache.items_by_type[
                "requirement/non-functional/performance-runtime"]:
            name_to_uid[item.ident] = item.uid
        perf_limits = item_cache[args.perf_limits_uid[0]]
        limits_by_uid: _LimitsByUID = {}
        for link in perf_limits["links"]:
            if link["role"] == "performance-runtime-limits":
                limits_by_uid[link["uid"]] = link["limits"]
        for report in args.reports:
            with open(report, "r", encoding="utf-8") as src:
                data = json.load(src)
            _update_perf_limits(data["reports"], name_to_uid, limits_by_uid,
                                args)
        _write_perf_limits(perf_limits, limits_by_uid)



# SPDX-License-Identifier: BSD-2-Clause
""" Builds an Software Verification Report (SVR). """

# Copyright (C) 2022, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import itertools
from typing import Iterator

from specitems import Item, ItemGetValueContext, SphinxContent

from .directorystate import DirectoryState
from .membench import generate
from .pkgitems import PackageBuildDirector
from .linkhub import LinkHub
from .rtems import gather_test_suites
from .sphinxbuilder import spacify
from .specdocbuilder import SpecDocumentBuilder
from .testaggregator import get_test_result_status, TestAggregator

_DEFINE_NOT_DEFINED = "requirement/functional/interface-define-not-defined"

_GROUP = ["interface/group", "requirement/non-functional/design-group"]


def _gather_design_components(item: Item, components: list[Item]) -> None:
    if item.type in _GROUP:
        components.append(item)
        return
    for parent in item.parents("interface-function"):
        components.append(parent)
    for parent in item.parents("requirement-refinement"):
        if parent.type in _GROUP:
            components.append(parent)
    if not components:
        for parent in item.parents("requirement-refinement"):
            _gather_design_components(parent, components)


def _get_name(kind_name: str) -> str:
    return kind_name[kind_name.find("/") + 1:]


def _get_kind_name(kind_name: str) -> tuple[str, str]:
    separator = kind_name.find("/")
    return kind_name[0:separator], kind_name[separator + 1:]


class SVRBuilder(SpecDocumentBuilder):
    """ Builds an Software Verification Report (SVR). """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        my_type = self.item.type
        self.mapper.add_get_value(
            f"{my_type}:/traceability-requirements-to-design",
            self._traceability_requirements_to_design)
        self.mapper.add_get_value(
            f"{my_type}:/traceability-design-to-requirements",
            self._traceability_design_to_requirements)
        self.mapper.add_get_value(f"{my_type}:/traceability-design-to-code",
                                  self._traceability_design_to_code)
        self.mapper.add_get_value(f"{my_type}:/traceability-code-to-design",
                                  self._traceability_code_to_design)
        self.mapper.add_get_value(f"{my_type}:/unit-verification",
                                  self._unit_verification)
        self.mapper.add_get_value(f"{my_type}:/code-coverage-achievement",
                                  self._code_coverage_achievement)
        self.mapper.add_get_value(f"{my_type}:/performance-summary",
                                  self._performance_summary)
        self.mapper.add_get_value(f"{my_type}:/memory-benchmarks",
                                  self._get_membench)
        self._design_to_code: list[tuple[str, str]] = []

    def _req_to_design(self) -> Iterator[tuple[Item, list[Item]]]:
        for item in self.spec.get_related_requirements():
            components: list[Item] = []
            _gather_design_components(item, components)
            yield (item, components)

    def _traceability_requirements_to_design(self,
                                             _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        rows = [["Requirement", "Design Component"]]
        for item, components in self._req_to_design():
            req = self.mapper.get_link(item)
            for component in sorted(set(components)):
                try:
                    link = component.view["document-links"]["sdd"]
                except KeyError:
                    group = "requirement/non-functional/design-group"
                    if item.type == _DEFINE_NOT_DEFINED:
                        link = "N/A (interface define is not defined)"
                    elif item.type == group and item["identifier"] is None:
                        link = "N/A (external design)"
                    else:
                        link = "**no reference to SDD**"
                rows.append([req, link])
                req = ""
            if req:
                rows.append(
                    [req, "N/A (no directly associated design components)"])
        content.add_grid_table(rows, [50, 50], font_size=-4)
        return str(content)

    def _traceability_design_to_requirements(self,
                                             _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        link_hub = self.input("link-hub")
        assert isinstance(link_hub, LinkHub)
        rows = [["Design Component", "Requirement"]]
        for _, info in sorted(link_hub.name_info.items(),
                              key=lambda x: _get_name(x[0])):
            if not isinstance(info, dict):
                continue
            variants = info.get("variants", [])
            for variant in itertools.chain([info], variants):
                name = f"`{spacify(variant['name'])} <{variant['link']}>`__"
                if variants:
                    file_path = variant["file"]
                    file_link = link_hub.name_info[f"file/{file_path}"]["link"]
                    name = f"{name} in `{spacify(file_path)} <{file_link}>`__"
                for item in sorted(variant["items"]):
                    url = item.view["document-links"]["url"]
                    rows.append([name, f"`{item.spec_2} <{url}>`__"])
        content.add_grid_table(rows, [60, 40], font_size=-4)
        return str(content)

    def _get_design_to_code(self) -> list[tuple[str, str]]:
        if self._design_to_code:
            return self._design_to_code
        link_hub = self.input("link-hub")
        assert isinstance(link_hub, LinkHub)
        design_to_code: list[tuple[str, str]] = []
        for kind_name, info in sorted(link_hub.name_info.items(),
                                      key=lambda x: _get_name(x[0])):
            if not isinstance(info, dict):
                continue
            kind, name = _get_kind_name(kind_name)
            if kind == "file":
                continue
            if kind == "group":
                name = f"`{spacify(info['name'])} <{info['link']}>`__"
                for file_path in info["files"]:
                    file_link = link_hub.name_info[f"file/{file_path}"]["link"]
                    file = f"`{spacify(file_path)} <{file_link}>`__"
                    design_to_code.append((name, file))
                continue
            variants = info.get("variants", [])
            for variant in itertools.chain([info], variants):
                name = f"`{spacify(variant['name'])} <{variant['link']}>`__"
                file_path = variant["file"]
                file_link = link_hub.name_info[f"file/{file_path}"]["link"]
                file = f"`{spacify(file_path)} <{file_link}>`__"
                design_to_code.append((name, file))
        self._design_to_code = design_to_code
        return design_to_code

    def _traceability_design_to_code(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        rows = [["Design Component", "File"]]
        last = ""
        for design, code in self._get_design_to_code():
            rows.append(["" if last == design else design, code])
            last = design
        content.add_grid_table(rows, [60, 40], font_size=-4)
        return str(content)

    def _traceability_code_to_design(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        code_to_design = sorted(
            (code, design) for design, code in self._get_design_to_code())
        rows = [["File", "Design Component"]]
        last = ""
        for code, design in code_to_design:
            rows.append(["" if last == code else code, design])
            last = code
        content.add_grid_table(rows, [40, 60], font_size=-4)
        return str(content)

    def _unit_verification(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        test_suites: list[Item] = []
        gather_test_suites(self.item.cache["/testsuites/unit"], test_suites)
        rows = [["Test Case", "Status"]]
        unspecified: dict[str, list[str]] = {}
        for test_suite in sorted(test_suites):
            for test_case in test_suite.children("test-case"):
                rows.append([
                    self.mapper.get_link(test_case),
                    get_test_result_status(test_case, "", "")
                ])
            for test_results in test_suite.view.get("test-results",
                                                    {}).values():
                for data in test_results:
                    for test_case in data["unspecified-test-cases"]:
                        unspecified.setdefault(test_case["name"], []).append(
                            f"`{data['status']} <{data['link']}>`__")
        rows_2 = [["Test Case without Specification", "Status"]]
        for name, status in sorted(unspecified.items()):
            rows_2.append([name, ", ".join(status)])
        content.add_grid_table(rows, [80, 20], font_size=-1)
        content.add_grid_table(rows_2, [80, 20], font_size=-1)
        return str(content)

    def _code_coverage_achievement(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        test_aggregator = self.input("test-aggregation")
        assert isinstance(test_aggregator, TestAggregator)
        test_aggregator.add_coverage_achievement(content)
        return str(content)

    def _performance_summary(self, _ctx: ItemGetValueContext) -> str:
        content = SphinxContent(section_level=2)
        types = ["requirement/non-functional/performance-runtime"]
        rows = [["Requirement", "Status"]]
        for item in self.spec.get_related_items_by_type(types):
            rows.append([
                self.mapper.get_link(item),
                get_test_result_status(item, "", "")
            ])
        content.add_grid_table(rows, [80, 20], font_size=-1)
        return str(content)

    def _get_membench(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            content = SphinxContent(section_level=self.section_level)
            link, membench = self.input_link("membench-results")
            assert isinstance(membench, DirectoryState)
            label = membench.substitute(link["build-label"])
            sections_by_uid = membench.json_load()[label]["membench"]
            root = self.item.cache["/rtems/req/mem-basic"]
            table_pivots = ["/rtems/req/mem-smp-1"]
            generate(content, sections_by_uid, root, table_pivots, self.mapper)
            return "\n".join(content)



# SPDX-License-Identifier: BSD-2-Clause
""" Provides utility functions. """

# Copyright (C) 2020, 2023 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import argparse
import logging
import os
import shutil

from specitems import create_argument_parser, ItemMapper


def duration(value: float | str) -> str:
    """ Convert a duration in seconds into a value with unit string. """
    if isinstance(value, str):
        return value
    assert value >= 0.0
    if value == 0.0:
        return "0s"
    if value < 1e-6:
        return f"{value * 1e9:.3f}ns"
    if value < 1e-3:
        return f"{value * 1e6:.3f}μs"
    if value < 1.0:
        return f"{value * 1e3:.3f}ms"
    return f"{value:.3f}s"


def copy_file(src_file: str, dst_file: str, log_context: str) -> None:
    """ Copy the source file to the destination file. """
    os.makedirs(os.path.dirname(dst_file), exist_ok=True)
    logging.info("%s: copy '%s' to '%s'", log_context, src_file, dst_file)
    shutil.copy2(src_file, dst_file)


def copy_files(src_dir: str, dst_dir: str, files: list[str],
               log_context: str) -> None:
    """
    Copy a list of files in the source directory to the destination directory
    preserving the directory of the files relative to the source directory.
    """
    for a_file in files:
        src_file = os.path.join(src_dir, a_file)
        dst_file = os.path.join(dst_dir, a_file)
        os.makedirs(os.path.dirname(dst_file), exist_ok=True)
        logging.info("%s: copy '%s' to '%s'", log_context, src_file, dst_file)
        shutil.copy2(src_file, dst_file)


def copy_and_substitute(src_file: str, dst_file: str, mapper: ItemMapper,
                        log_context: str) -> None:
    """
    Copy the file from the source to the destination path and performs a
    variable substitution on the file content using the item mapper.
    """
    logging.info("%s: read: %s", log_context, src_file)
    with open(src_file, "r", encoding="utf-8") as src:
        logging.info("%s: substitute using mapper of item %s", log_context,
                     mapper.item.uid)
        content = mapper.substitute(src.read())
        logging.info("%s: write: %s", log_context, dst_file)
        os.makedirs(os.path.dirname(dst_file), exist_ok=True)
        with open(dst_file, "w+", encoding="utf-8") as dst:
            dst.write(content)


def remove_empty_directories(scope: str, base: str) -> None:
    """
    Recursively remove all empty subdirectories of base and base itself if it
    gets empty.

    The scope is used for log messages.
    """
    removed: set[str] = set()
    for root, subdirs, files in os.walk(base, topdown=False):
        if files:
            continue
        if any(subdir for subdir in subdirs
               if os.path.join(root, subdir) not in removed):
            continue
        logging.info("%s: remove empty directory: %s", scope, root)
        os.rmdir(root)
        removed.add(root)


def create_build_argument_parser(
        default_log_level: str = "INFO") -> argparse.ArgumentParser:
    """ Create an argument parser with default build options. """
    parser = create_argument_parser(default_log_level)
    parser.add_argument('--only',
                        type=str,
                        action="append",
                        default=None,
                        help="build only these steps")
    parser.add_argument('--force',
                        type=str,
                        action="append",
                        default=None,
                        help="force to build these steps")
    parser.add_argument('--skip',
                        type=str,
                        action="append",
                        default=None,
                        help="skip these steps")
    parser.add_argument('--no-spec-verify',
                        action="store_true",
                        help="do not verify the specification")
    return parser



# SPDX-License-Identifier: BSD-2-Clause
"""
Provides a command line interface to update the test timeouts using the
specified test reports.
"""

# Copyright (C) 2024, 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import argparse
import contextlib
import json
import logging
from pathlib import Path
import sys

from specitems import (ItemCache, ItemCacheConfig, create_argument_parser,
                       create_config, init_logging, item_is_enabled)
from specware import load_specware_config, SpecWareTypeProvider


def _update_timeouts(args: argparse.Namespace, report_path: str,
                     timeouts: dict[str, list[int]], reports: list) -> None:
    error_factor = args.error_factor
    warning_factor = args.warning_factor
    minimum_timeout = args.minimum_timeout
    for report in reports:
        name = Path(report["executable"]).name
        new_duration = report.get("execution-duration-in-seconds",
                                  report["duration"])
        if new_duration == 0.0:
            continue
        durations = timeouts.setdefault(name, [])
        try:
            maximum = max(durations)
        except ValueError:
            pass
        else:
            if new_duration > error_factor * maximum + minimum_timeout:
                logging.error(
                    "%s: %s: duration %s is greater than %s * %s + %s",
                    report_path, name, new_duration, error_factor, maximum,
                    minimum_timeout)
                continue
            if new_duration > warning_factor * maximum + minimum_timeout:
                logging.warning(
                    "%s: %s: duration %s is greater than %s * %s + %s",
                    report_path, name, new_duration, warning_factor, maximum,
                    minimum_timeout)
        durations.append(new_duration)


def cliupdatetimeouts(argv: list[str] = sys.argv) -> None:
    """ Update the test timeouts using the specified test reports. """
    parser = create_argument_parser()
    parser.add_argument("--config-file",
                        type=str,
                        default=None,
                        help="use this configuration file")
    parser.add_argument('--dry-run',
                        action="store_true",
                        default=False,
                        help="do not save timeout items")
    parser.add_argument('--reset',
                        action="store_true",
                        help="reset the timeouts")
    parser.add_argument('--test-timeouts',
                        type=str,
                        default="test-timeouts",
                        help="the test timeouts item UID basename")
    parser.add_argument('--minimum-timeout',
                        type=float,
                        default=10.0,
                        help="the minimum timeout used by the test runner")
    parser.add_argument(
        '--warning-factor',
        type=float,
        default=1.2,
        help="a new duration greater than the factor times "
        "the current maximum duration plus the minimum timeout is a warning")
    parser.add_argument(
        '--error-factor',
        type=float,
        default=1.9,
        help="a new duration greater than the factor times "
        "the current maximum duration plus the minimum timeout is an error")
    parser.add_argument("reports",
                        metavar="REPORTS",
                        nargs="+",
                        help="the test report files")
    args = parser.parse_args(argv[1:])
    init_logging(args)
    config, working_directory = load_specware_config(args.config_file)
    with contextlib.chdir(working_directory):
        item_cache = ItemCache(create_config(config["spec"], ItemCacheConfig),
                               type_provider=SpecWareTypeProvider({}),
                               is_item_enabled=item_is_enabled)
        reset_done: set[str] = set()
        for report_path in args.reports:
            logging.info("%s: evaluate reports", report_path)
            with open(report_path, "r", encoding="utf-8") as src:
                data = json.load(src)
            try:
                target = data['target']
            except KeyError:
                continue
            timeout_key = data["timeout-key"]
            item = item_cache[f"{target}/{args.test_timeouts}"]
            reset_key = f"{item.uid} {timeout_key}"
            if args.reset and reset_key not in reset_done:
                reset_done.add(reset_key)
                timeouts = {}
            else:
                timeouts = item["timeouts"].get(timeout_key, {})
            item["timeouts"][timeout_key] = timeouts
            _update_timeouts(args, report_path, timeouts, data["reports"])
            if not args.dry_run:
                item.save()



# SPDX-License-Identifier: BSD-2-Clause
""" Provides a test runner using the ESA run tests service.  """

# Copyright (C) 2023 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import bz2
import datetime
import logging
import math
from pathlib import Path
import socket
import time
from typing import Dict, List, Tuple
import yaml

from specware import run_command

from .directorystate import RepositoryState
from .testrunner import Executable, Report, TestRunner


def _bz2_file(executable: Executable) -> str:
    return f"{Path(executable.path).name}.bz2"


class ESATestRunner(TestRunner):
    """
    Runs the tests using the ESA run tests service.

    The interface to the service is a Git repository located at:

    https://gitrepos.estec.esa.int/ttsiodras/RemoteExecutionOfTestBinaries
    """

    def describe(self) -> str:
        return f"""This test procedure requests the `ESA run tests service
<https://gitrepos.estec.esa.int/ttsiodras/RemoteExecutionOfTestBinaries>`__ to
run the test programs for the ``{self.component['bsp'].upper()}`` target
board."""

    def _create_branch(self, working_directory: Path,
                       repository: RepositoryState) -> str:
        """
        Create a unique branch to carry out the request.

        Returns the branch name.
        """
        hostname = socket.gethostname()
        now = datetime.datetime.utcnow().strftime('%Y-%m-%d-%H-%M-%S-%f')
        branch = f"pkg-{self.component['name']}-{hostname}-{now}"
        status = run_command(
            ["git", "checkout", "-B", branch, repository["commit"]],
            cwd=working_directory)
        assert status == 0
        return branch

    def _copy_and_prepare_executables(self, working_directory: Path,
                                      executables: List[Executable]) -> None:
        """
        Copy, strip, and compress (bzip2) the executables.

        The files are placed into the working directory.
        """
        strip = self["strip-program-path"]
        for executable in executables:
            source = Path(executable.path)
            stripped = working_directory / source.name
            status = run_command(
                [strip, str(source), "-o",
                 str(stripped)],
                cwd=working_directory)
            assert status == 0
            with stripped.open("rb") as src:
                data = bz2.compress(src.read())
            stripped.unlink()
            compressed = working_directory / _bz2_file(executable)
            with compressed.open("wb") as dst:
                dst.write(data)

    def _write_request_file(self, working_directory: Path,
                            executables: List[Executable]) -> Tuple[str, int]:
        """
        Write the request file to run the prepared executables.

        Returns the request file name and the overall timeout in seconds.
        """
        request_file = working_directory / "request.yaml"
        overall_timeout = 0
        with request_file.open("w", encoding="utf-8") as dst:
            jobs: List[Dict[str, Dict[str, int]]] = []
            max_timeout = 0
            for executable in executables:
                timeout = int(math.ceil(executable.timeout))
                overall_timeout += timeout
                max_timeout = max(max_timeout, timeout)
                jobs.append(
                    {_bz2_file(executable): {
                         "timeout_in_seconds": timeout
                     }})
            request = {
                "version": 1,
                "target_board": self.component["bsp"].upper(),
                "timeout_in_seconds": max_timeout,
                "jobs": jobs
            }
            dst.write(
                yaml.dump(request,
                          default_flow_style=False,
                          allow_unicode=True))
        return request_file.name, overall_timeout

    def _wait_for_update(self, working_directory: Path, branch: str,
                         overall_timeout: int) -> None:
        """
        Wait for a branch update on the remote repository.

        The time to wait is limited by the maximum request delay and the
        overall timeouts of the executables.
        """
        max_duration = float(self["max-request-delay-in-seconds"] +
                             overall_timeout)
        begin = time.monotonic()
        while time.monotonic() - begin < max_duration:
            time.sleep(self["git-fetch-delay-in-seconds"])
            status = run_command(["git", "fetch", "origin", branch],
                                 cwd=working_directory)
            if status != 0:
                logging.warning("%s: fetch from remote repository failed",
                                self.uid)
            stdout: List[str] = []
            status = run_command(
                ["git", "log", f"HEAD..origin/{branch}", "--oneline"],
                stdout=stdout,
                cwd=working_directory)
            assert status == 0
            if stdout:
                return

    def _get_reports(self, working_directory: Path,
                     executables: List[Executable]) -> List[Report]:
        """ Get the reports from the result text file of each executable.  """
        reports: List[Report] = []
        for executable in executables:
            result_file = f"{Path(executable.path).name}.bz2.result.txt"
            result_path = working_directory / result_file
            report: Report = {
                "executable": executable.path,
                "executable-sha512": executable.digest,
                "command-line": []
            }
            try:
                with result_path.open("r", encoding="utf-8") as src:
                    output = src.read().rstrip().replace("\r\n",
                                                         "\n").split("\n")
            except FileNotFoundError:
                continue
            report["output"] = output
            reports.append(report)
        return reports

    def run_tests(self, executables: List[Executable]) -> List[Report]:
        super().run_tests(executables)
        repository = self.input("repository")
        assert isinstance(repository, RepositoryState)
        working_directory = Path(repository.directory)
        branch = self._create_branch(working_directory, repository)
        self._copy_and_prepare_executables(working_directory, executables)
        request_file, overall_timeout = self._write_request_file(
            working_directory, executables)
        files = [_bz2_file(executable) for executable in executables]
        status = run_command(["git", "add"] + files + [request_file],
                             cwd=working_directory)
        assert status == 0
        status = run_command(["git", "commit", "-m", "Request"],
                             cwd=working_directory)
        assert status == 0
        status = run_command(["git", "push", "-u", "origin", branch],
                             cwd=working_directory)
        assert status == 0
        self._wait_for_update(working_directory, branch, overall_timeout)
        status = run_command(["git", "reset", "--hard", f"origin/{branch}"],
                             cwd=working_directory)
        assert status == 0
        status = run_command(["git", "push", "-d", "origin", branch],
                             cwd=working_directory)
        assert status == 0
        reports = self._get_reports(working_directory, executables)
        status = run_command(["git", "checkout", repository["branch"]],
                             cwd=working_directory)
        assert status == 0
        status = run_command(["git", "branch", "-D", branch],
                             cwd=working_directory)
        assert status == 0
        return reports
